import sys

class HTTPError (Exception):
    def __init__ (self, status = "200 OK", explain = "", errno = 0, traceback = False):
        self.status = status
        self.explain = explain
        self.errno = errno
        self.exc_info = traceback and sys.exc_info () or None



from . import wsgi_handler, proxy_handler
from aquests.protocols.http import request as http_request
import re
from rs4.termcolor import tc

class Handler (proxy_handler.Handler):
	def __init__ (self, wasc, clusters, cachefs = None):
		proxy_handler.Handler.__init__ (self, wasc, clusters, cachefs)
		self.route_map = {}
		self.sorted_route_map = []

	def add_route (self, route, cname):
		try:
			cname, mapped_path = cname.split ("/", 1)
		except ValueError:
			cname, mapped_path = cname, ''
		else:
			cname, mapped_path = cname, "/" + mapped_path

		self.wasc.logger ('server', 'upstream @{} mounted to {}'.format (cname, tc.white (route)), 'info')
		self.route_map [route] = (cname, len (route), mapped_path)
		temp = list (self.route_map.items ())
		temp.sort (key = lambda x: x [1][1], reverse = True)
		self.sorted_route_map = temp

	def match (self, request):
		return self.find_cluster (request) and 1 or 0

	def find_cluster (self, request):
		uri = request.uri
		for route, (cname, route_len, mapped_path) in self.sorted_route_map:
			if uri.startswith (route):
				return self.clusters [cname], route_len, mapped_path

	def will_open_tunneling (self):
		return self.response.code == 101 # websocket connection upgrade

	def continue_request (self, request, collector):
		request.set_header ("X-GTXN-ID", request.gtxid)
		request.set_header ("X-LTXN-ID", request.ltxid)
		request.loadbalance_retry = 0
		if self.has_valid_cache (request, collector is not None):
			return

		try:
			self.route (request, collector)
		except:
			self.wasc.logger.trace ("server")
			request.response.error (500, "", "Routing failed. Please contact administator.")

	def route (self, request, collector):
		current_cluster, route_len, mapped_path  = self.find_cluster (request)
		psysicaluri = request.uri [route_len:]
		if psysicaluri == "": psysicaluri = "/"
		elif psysicaluri[0] != "/": psysicaluri = "/" + psysicaluri
		if mapped_path:
			psysicaluri = mapped_path + psysicaluri

		request.loadbalance_retry += 1
		retry = request.loadbalance_retry
		asyncon = current_cluster.get (index = -1)

		if not asyncon:
			request.logger ("nopool-%d, url: %s, cluster index: %d" % (
				retry,
				request.uri,
				route
				), 'warn')
			return request.response.error (503)

		if retry > 1:
			request.logger ("failsafe-%d, target: %s:%s, url: %s, cluster index: %d" % (
				retry,
				asyncon.address [0],
				asyncon.address [1],
				request.uri,
				route
			), 'warn')

		fetcher = http_request.HTTPRequest (psysicaluri, request.command, collector is not None, logger = self.wasc.logger.get ("server"))
		r = proxy_handler.proxy_request_handler (asyncon, fetcher, self.callback, request, collector)
		r.handle_request ()

	def callback (self, handler):
		request, response, collector = handler.client_request, handler.response, handler.collector
		cluster = self.find_cluster (request) [0]
		cluster.report (handler.asyncon, response.code)

		if response.code >= 700:
			if request.loadbalance_retry >= len (cluster):
				request.response.error (506, response.msg)

			elif request.channel:
				if not collector or (collector and collector.cached):
					if collector:
						collector.reuse_cache ()
					self.route (request, collector)
					return self.dealloc (request, handler)

		else:
			try:
				self.save_cache (request, handler)
			except:
				self.wasc.logger.trace ("server")

		request.loadbalance_retry = 0
		self.dealloc (request, handler)



# Subprocess Task

from . import task
import subprocess
from concurrent.futures import TimeoutError
from ..tasks import Mask
import time
from skitai import was

class Task (task.Task):
    def __init__ (self, cmd, meta, filter = None, timeout = None):
        self.setup (cmd, meta, filter, timeout)
        self.proc = subprocess.Popen (cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell = True)

    def __getattr__ (self, name):
        raise AttributeError

    @property
    def lines (self):
        for line in iter (self.proc.stdout.readline, b''):
            yield line

    def then (self, func):
        self._fulfilled = func
        self._was = self._get_was ()
        return was.Thread (self._settle)

    def _settle (self, future = None):
        if self._fulfilled:
            mask = self._create_mask (0)
            if '__reqid' in self._meta:
                self._fulfilled (mask)
                self._fulfilled = None
            else:
                self._late_respond (mask)

    def set_callback (self, func, reqid = None, timeout = None):
        if reqid is not None:
            self._meta ["__reqid"] = reqid
        self._timeout = timeout
        self._fulfilled = func
        was.Thread (self._settle)

    def kill (self):
        self.proc.kill ()

    def terminate (self):
        self.proc.terminate ()

    def _create_mask (self, timeout):
        if timeout:
            self._timeout = timeout
        if self._mask:
            return self._mask

        data, expt = None, None
        try:
            data, err = self.proc.communicate (timeout = self._timeout)
        except subprocess.TimeoutExpired:
            expt = TimeoutError
            self.proc.terminate ()
            self.proc.wait ()
        else:
            if self.proc.returncode:
                if isinstance (err, bytes):
                    err = err.decode ()
                expt = SystemError ('code:{} {}'.format (self.proc.returncode, err))
        if isinstance (data, bytes):
            data = data.decode ()
        if self._filter:
            data =  self._filter (data)
        self._mask = Mask (data, expt, meta = self.meta)
        return self._mask



import json
UJSON = False

import sys
from xmlrpc import client
from ..utility import catch
import copy
from datetime import timezone, datetime, date
from ..exceptions import HTTPError

TZ_LOCAL = datetime.now (timezone.utc).astimezone().tzinfo
TZ_UTC = timezone.utc

class BaseDecoder:
	def _to_utc (self, obj):
		try:
			return obj.astimezone (TZ_UTC)
		except ValueError:
			return obj.replace (tzinfo = TZ_LOCAL).astimezone (TZ_UTC)

if UJSON:
	class StringDateTimeEncoder (BaseDecoder):
		def default (self, obj):
			if isinstance (obj, date):
				return str (obj)
			return obj
else:
	class StringDateTimeEncoder (json.JSONEncoder, BaseDecoder):
		def default (self, obj):
			if isinstance (obj, date):
				return str (obj)
			return super ().default (obj)

class DigitTimeEncoder (StringDateTimeEncoder):
	def default (self, obj):
		if isinstance (obj, date):
			try: return self._to_utc (obj).strftime ('%Y%m%d%H%M%S')
			except AttributeError: return obj.isoformat ()
		return obj if UJSON else super ().default (obj)

class ISODateTimeEncoder (StringDateTimeEncoder):
	def default (self, obj):
		if isinstance (obj, date):
			try: return self._to_utc (obj).isoformat ()
			except AttributeError: return obj.isoformat ()
		return obj if UJSON else super ().default (obj)

class UNIXEpochDateTimeEncoder (StringDateTimeEncoder):
	def default (self, obj):
		if isinstance (obj, date):
			try: return self._to_utc (obj).timestamp ()
			except AttributeError: return obj.isoformat ()
		return obj if UJSON else super ().default (obj)

class ISODateTimeWithOffsetEncoder (StringDateTimeEncoder):
	def default (self, obj):
		if isinstance (obj, date):
			try: return self._to_utc (obj).strftime ('%Y-%m-%d %H:%M:%S+00')
			except AttributeError: return obj.isoformat ()
		return obj if UJSON else super ().default (obj)


ENCODER_MAP = {
	'str': StringDateTimeEncoder,
	'iso': ISODateTimeEncoder,
	'unixepoch': UNIXEpochDateTimeEncoder,
	'digit': DigitTimeEncoder,
	'utcoffset': ISODateTimeWithOffsetEncoder
}
if UJSON:
	for k, v in list (ENCODER_MAP.items ()):
		ENCODER_MAP [k] = v ()

def tojson (data, pretty = False):
	if UJSON:
		return json.dumps (data, ensure_ascii = False, indent = pretty and 2 or 0, pre_encode_hook = ENCODER_MAP ['utcoffset'].default)
	else:
		return json.dumps (data, ensure_ascii = False, indent = pretty and 2 or None, cls = ENCODER_MAP ['utcoffset'])

class APIResponse:
	def __init__ (self, content_type):
		self.content_type = self.set_content_type (content_type)
		self.data_encoder_class = None
		self.pretty = False
		self.data = {}

	def __call__ (self, __do_not_use_variable__ = None, **kargs):
		self.data = __do_not_use_variable__ or kargs
		return self.to_string ()

	def set_content_type (self, content_type):
		if content_type.find ("text/xml") != -1:
			return "text/xml"
		return 'application/json'

	def set_json_encoder (self, encoder, pretty = True):
		self.pretty = pretty
		if not encoder:
			return
		if isinstance (encoder, str):
			self.data_encoder_class = ENCODER_MAP [encoder]
			return
		self.data_encoder_class = encoder

	def get_content_type (self):
		return self.content_type

	def __setitem__ (self, k, v):
		self.set (k, v)

	def __getitem__ (self, k, v = None):
		return self.get (k, v)

	def set (self, k, v):
		self.data [k] = v

	def get (self, k, v = None):
		return self.data.get (k, v)

	def encode (self, charset):
		return self.to_string ().encode (charset)

	def __str__ (self):
		return self.to_string ()

	def to_string (self):
		if self.content_type.startswith ("text/xml"):
			return client.dumps ((self.data,))

		if UJSON:
			return json.dumps (self.data, ensure_ascii = False, indent = self.pretty and 2 or 0, pre_encode_hook = (self.data_encoder_class or ENCODER_MAP ['utcoffset']).default)
		else:
			return json.dumps (self.data, ensure_ascii = False, indent = self.pretty and 2 or None, cls = self.data_encoder_class or ENCODER_MAP ['utcoffset'])

	def traceback (self, message = 'exception occured', code = 20000, debug = 'see traceback', more_info = None):
		self.error (message, code, debug, more_info, sys.exc_info ())

	def error (self, message, code = 20000, debug = None, more_info = None, exc_info = None):
		self.data = {}
		self.data ['code'] = code
		self.data ['message'] = message
		if debug:
			self.data ['debug'] = debug
		if more_info:
			self.data ['more_info'] = more_info
		if exc_info:
			self.data ["traceback"] = type (exc_info) is tuple and catch (2, exc_info) or exc_info

	def set_spec (self, app):
		pass


class API (APIResponse):
	@classmethod
	def add_custom_encoder (cls, name, encoder):
		ENCODER_MAP [name] = encoder

	def __init__ (self, request, data = None):
		self.request = request # for response typing
		self.data = data or {}
		self.content_type = self.set_content_type (self.request.get_header ("accept", ''))
		self.data_encoder_class = None
		self.pretty = False

	def set_spec (self, app):
		resource_id = self.request.routable ["func_id"]
		routable = copy.deepcopy (self.request.routable)
		del routable ["func_id"]
		routable ["methods"] = list (routable ["methods"])
		props = {"current_request": {}}
		props ["current_request"]["method"] = self.request.method
		props ["current_request"]["version"] = self.request.version
		props ["current_request"]["uri"] = self.request.uri
		props ['routeopt'] = routable
		props ["parameter_requirements"] = app.get_parameter_requirements (resource_id)
		props ['auth_requirements'] = app.get_auth_flags (resource_id)
		props ["doc"] = self.request.routed.__doc__
		props ["id"] = resource_id
		self.data ["__spec__"] = props



"""
Hans Roh 2015 -- http://osp.skitai.com
License: BSD
"""
import re
import sys
import os
import shutil, glob
import codecs
from warnings import warn
try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup

with open('skitai/__init__.py', 'r') as fd:
	version = re.search(r'^__version__\s*=\s*"(.*?)"',fd.read(), re.M).group(1)

classifiers = [
  'License :: OSI Approved :: MIT License',
  'Development Status :: 4 - Beta',
  'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
	'Topic :: Internet :: WWW/HTTP :: WSGI',
	'Environment :: Console',
	'Environment :: No Input/Output (Daemon)',
	'Topic :: Internet',
	'Topic :: Software Development :: Libraries :: Python Modules',
	'Intended Audience :: Developers',
	'Programming Language :: Python :: 3.5',
	'Programming Language :: Python :: 3.6',
	'Programming Language :: Python :: 3.7',
	'Programming Language :: Python :: Implementation :: CPython',
	'Programming Language :: Python :: Implementation :: PyPy'
]

packages = [
	'skitai',
	'skitai.scripts',
	'skitai.scripts.commands',
	'skitai.backbone',
	'skitai.handlers',
	'skitai.handlers.http2',
	'skitai.handlers.websocket',
	'skitai.handlers.proxy',
	'skitai.corequest',
	'skitai.corequest.dbi',
	'skitai.corequest.httpbase',
	'skitai.corequest.pth',
	'skitai.wastuff',
	'skitai.mounted',
	'skitai.testutil'
]

package_dir = {'skitai': 'skitai'}
package_data = {}

install_requires = [
	"rs4>=0.2.5.0",
	"aquests>=0.29.2.0",
	"jsonrpclib-pelix",
	"sqlphile",
	"pebble",
]
if os.name == "nt":
	install_requires.append ("pywin32")

with codecs.open ('README.rst', 'r', encoding='utf-8') as f:
	long_description = f.read()

setup (
	name='skitai',
	version=version,
	description='Skitai App Engine',
	long_description=long_description,
	url = 'https://gitlab.com/hansroh/skitai',
	author='Hans Roh',
	author_email='hansroh@gmail.com',
	packages=packages,
	package_dir=package_dir,
	package_data = package_data,
	entry_points = {
      'console_scripts': [
			'skitai=skitai.scripts.skitai:main',
		],
    },
	license='MIT',
	platforms = ["posix", "nt"],
	download_url = "https://pypi.python.org/pypi/skitai",
	install_requires = install_requires,
	classifiers=classifiers
)
